### Produces Table 4 ("Simulation results of CRR and Poisson with redemption and recover")
### One must change the value of the "brownian" variable to switch between CRR and Poisson

library(stats)
library(statmod)
library(parallel)
library(mFilter)

h <- 1
brownian <- TRUE
min.a <- -0.5

### Brownian and Poisson observationally equivalent at h=h0

sigma <- 0.022
mu <- 0.01
h0 <- 4

### 1/(4*sigma+mu/sigma)^2 = 3.40

num.cores <- detectCores()

  gam <- 2
  lambda <- 0.01
  beta <- 0.8^h
  r <- 1.01^h-1

  x <- 1/5/4
  recovery <- -0.2
  
  if (brownian) {
    gu <- exp(sigma*sqrt(h))
    gd <- 1/gu
    pu <- 0.5 + mu/2/sigma*sqrt(h)
    pd <- 1-pu
  } else {
    p0 <- -log(0.5 + mu/2/sigma*sqrt(h0))/h0
    
    gd <- exp(-sigma*sqrt(h0))/(p0*h0/(1-exp(-p0*h0)))*(p0*h/(1-exp(-p0*h)))
    pu <- exp(-p0*h)
    pd <- 1-exp(-p0*h)

    ## Compute gu so that pu*gu+pd*gd the same than in the Brownian of same h
    x.brownian <- (0.5 + mu/2/sigma*sqrt(h))*exp(sigma*sqrt(h)) + (0.5 - mu/2/sigma*sqrt(h))*exp(-sigma*sqrt(h))
    gu <- (x.brownian - pd*gd)/pu
  }

  stopifnot(pu > 0 && pd > 0)

  stopifnot(beta*(pu*gu^(1-gam) + pd*gd^(1-gam))*(1+r) < 1)

  max.a <- 0
  ngrid.a <- 200
  convergence.tol <- 1e-5

  VG <- NULL
  VB <- NULL
  threshold <- NA

  ## Grids
  grid.a <- seq(min.a, max.a, length.out = ngrid.a)

  util <- function(c)
    ifelse(c <= 0, -1e8, c^(1-gam)/(1-gam))

  V <- function(a)
    max(VB, VG(a))

  q <- function(a1)
    {
      if (a1 >= 0)
        return(1/(1+r))
      
      Erepay1 <- 0

      if (VG(a1/gu) >= VB)
        Erepay1 <- Erepay1 + pu

      if (VG(a1/gd) >= VB)
        Erepay1 <- Erepay1 + pd
      
      Erepay1/(1+r)
    }

  a1.dr <- function(a)
    {
      VG.objective <- function(a1)
        util(h+a-a1*q(a1)) + beta*(pu*gu^(1-gam)*V(a1/gu) + pd*gd^(1-gam)*V(a1/gd))
      
      zd <- max(min(threshold*gd, max.a), min.a)
      o1 <- optimize(VG.objective, lower = min.a, upper = zd, maximum = TRUE, tol = convergence.tol)
      o2 <- optimize(VG.objective, lower = zd, upper = max.a, maximum = TRUE, tol = convergence.tol)

      ifelse(o1$objective > o2$objective, o1$maximum, o2$maximum)
    }

  solve <- function(max.it = 10000)
    {
      ## Initial guess
      VB <<- util((1-lambda)*h)
      VG.points <- util(h + grid.a*r)
      threshold <<- min(max(-lambda*h/r, min.a), max.a)

      init.spline <- function()
        VG <<- splinefun(grid.a, VG.points, method = "monoH.FC")

      init.spline()

      ## Run iterations
      
      it <- 1
      err <- Inf

      VG.points.new <- numeric(ngrid.a)

      while(err > convergence.tol && it <= max.it)
        {
          difference <- function(a)
            VG(a)-VB

          VB.new <- util((1-lambda)*h) + beta*(pu*gu^(1-gam) + pd*gd^(1-gam))*(exp(-x*h)*VB + (1-exp(-x*h))*V(recovery))
          
          ## Compute VG
          VG.optimize <- function(a)
            {
              VG.objective <- function(a1)
                util(h+a-a1*q(a1)) + beta*(pu*gu^(1-gam)*V(a1/gu) + pd*gd^(1-gam)*V(a1/gd))
              
              zd <- max(min(threshold*gd, max.a), min.a)
              o1 <- optimize(VG.objective, lower = min.a, upper = zd, maximum = TRUE, tol = convergence.tol)
              o2 <- optimize(VG.objective, lower = zd, upper = max.a, maximum = TRUE, tol = convergence.tol)

              max(o1$objective, o2$objective)
            }
          VG.points.new <- as.numeric(mclapply(grid.a, VG.optimize, mc.cores = num.cores))

          ## Compute error
          err <- max(abs(VB.new-VB),
                     max(abs(VG.points.new-VG.points)))

          ## Copy
          VB <<- VB.new
          VG.points <- VG.points.new
          init.spline()

          ## Threshold
          if (difference(min.a) < 0 && difference(max.a) > 0)
            {
              threshold <<- uniroot(difference, lower = min.a, upper = max.a, tol = convergence.tol)$root
              cat(threshold, fill = TRUE)
            }
          else if (difference(min.a) >= 0)
            threshold <<- min.a
          else
            threshold <<- max.a

          ## Output
          cat("Iteration ", it, ": error=", err, fill = TRUE, sep = "")

          it <- it+1
        }

      if (threshold > a1.dr(threshold)/gd + convergence.tol)
        cat("Models with defaults", fill = TRUE)
      else
        cat("No defaults in this model", fill = TRUE)
      
      if (it == max.it+1)
        warning("Maximum iterations reached")

      threshold
    }

  simulate <- function(a0, d0, length)
    {
      a <- a0
      d <- d0 ## TRUE means "not in default"
      g <- NA
      cons <- numeric()
      spread <- numeric()
      tb.y <- numeric()

      draw.growth <- function() ifelse(runif(1) < pu, gu, gd) - 1
      
      for (i in 1:length)
        {
          if (d[i])
            if (VG(a[i]) >= VB)
              {
                d1 <- TRUE
                a1 <- a1.dr(a[i])
                cons0 <- h+a[i]-a1*q(a1)
                spread0 <- (1/q(a1)/(1+r))^(4/h)-1
                tb.y0 <- (h-cons0)/h
                g1 <- draw.growth()
                a1 <- a1/(1+g1)
              }
            else
              {
                d1 <- FALSE
                a1 <- NA
                cons0 <- h
                spread0 <- NA
                tb.y0 <- 0
                g1 <- draw.growth()
              }
          else
            {
              d1 <- runif(1) < x
              a1 <- ifelse(d1, recovery, NA)
              cons0 <- h
              spread0 <- NA
              tb.y0 <- 0
              g1 <- draw.growth()
            }
            
          a <- c(a, a1)
          d <- c(d, d1)
          g <- c(g, g1)
          cons <- c(cons, cons0)
          spread <- c(spread, spread0)
          tb.y <- c(tb.y, tb.y0)
        }

      cons <- c(cons, NA)
      spread <- c(spread, NA)
      tb.y <- c(tb.y, NA)

      list(a=a, d=d, g=g, cons=cons, spread=spread, tb.y=tb.y)
    }

  simulation.has.default <- function(a0, length)
    !all(simulate(a0, TRUE, length)$d)


  default.prob <- function(nsims, sim.length = 40/h)
    {
      count <- 0
      for (i in 1:nsims)
        if (simulation.has.default(0, sim.length))
          count <- count + 1
      count / nsims
    }
  
  threshold <- solve()

compute.gap <- function(x)
{
  x.no.na <- which(!is.na(x))
  x.hp <- hpfilter(x[x.no.na], freq=1600)
  x.gap <- x.hp$cycle
  x[x.no.na] <- x.gap
  x
}

sim.length <- 1000
warmup <- 15
nsims <- 500

as <- numeric()
ds <- logical()
gdp.gaps <- numeric()
cons.gaps <- numeric()
tb.y.gaps <- numeric()
spread.gaps <- numeric()

for (it in 1:nsims)
{
  l <- simulate(0,TRUE,sim.length)
  gdp <- cumprod(c(1,1+l$g[2:(sim.length+1)]))
  log.gdp <- log(gdp)
  log.cons <- log(l$cons * gdp)

  gdp.gap <- compute.gap(log.gdp)
  cons.gap <- compute.gap(log.cons)
  tb.y.gap <- compute.gap(l$tb.y)
  spread.gap <- compute.gap(l$spread)
  
  as <- c(as, l$a[(warmup+1):(sim.length+1)])
  ds <- c(ds, l$d[(warmup+1):(sim.length+1)])
  gdp.gaps <- c(gdp.gaps, gdp.gap[(warmup+1):(sim.length+1)])
  cons.gaps <- c(cons.gaps, cons.gap[(warmup+1):(sim.length+1)])
  tb.y.gaps <- c(tb.y.gaps, tb.y.gap[(warmup+1):(sim.length+1)])
  spread.gaps <- c(spread.gaps, spread.gap[(warmup+1):(sim.length+1)])
}

a.mean <- mean(as, na.rm = TRUE)

defs <- which(!ds[2:length(ds)] & ds[1:(length(ds)-1)])

d.mean <- 1-(1-length(defs)/length(ds))^(4/h)


cat(sprintf("Mean debt: %.1f\n", a.mean*100))
cat(sprintf("Default probability (annual): %.1f\n", d.mean*100))
cat(sprintf("sigma(C): %.1f\n", sd(cons.gaps, na.rm = TRUE)*100))
cat(sprintf("sigma(Q): %.1f\n", sd(gdp.gaps)*100))
cat(sprintf("sigma(TB/Q): %.1f\n", sd(tb.y.gaps, na.rm = TRUE)*100))
cat(sprintf("rho(TB/Q,Q): %.1f\n", cor(tb.y.gaps, gdp.gaps, use = "complete.obs")*100))
cat(sprintf("mean OG at default: %.1f\n", mean(gdp.gaps[defs])*100))
